from enbios2.base.unit_registry import ureg
from enbios2.bw2.extract_from_xml import parse_xml
from enbios2.generic.files import DataPath

# ureg.parse_expression("kWh")

# data_path: DataPath = DataPath("ecoinvent/ecoinvent 3.9.1_cutoff_ecoSpold02/MasterData/Units.xml")
# res = parse_xml(data_path, "unit", [], ["name", "comment"])
#
# for unit in res:
#     u = str(unit["name"])
#     try:
#         p = ureg.parse_expression(u)
#         print(u, " - ", p)
#     except Exception as e:
#         print(u, " ? ", unit["comment"])
#     finally:
#         print("******")

print(ureg.parse_expression('kilowatt hour'))
print(ureg.parse_expression('kilowatt hour') * 1)
print(ureg.parse_expression('kilowatt_hour') * 1)
from enbios2.base.unit_registry import ureg as o_ureg
print(o_ureg.parse_expression('kilowatt hour'))

# ureg.define('kilowatt_hour = hour * kilowatt')
print(o_ureg.parse_expression('kilowatt hour'))



from typing import Optional, Union

from numpy import ndarray
from pandas import DataFrame
from sklearn.preprocessing import MinMaxScaler

from enbios2.base.experiment import Experiment


class ResultsSelector:

    def __init__(self, experiment: Experiment,
                 scenarios: Optional[list[str]] = None,
                 methods: Optional[list[str]] = None):
        self.experiment = experiment

        all_aliases = [sc.alias for sc in self.experiment.scenarios]
        if scenarios is not None:
            for scenario in scenarios:
                if scenario not in all_aliases:
                    raise ValueError(f"Scenario {scenario} not found in experiment")
            self.scenarios = scenarios
        else:
            self.scenarios = all_aliases

        if methods is not None:
            for method in methods:
                if method not in self.experiment.methods.keys():
                    raise ValueError(f"Method {method} not found in experiment")
            self.methods = methods
        else:
            self.methods: list[str] = [m for m in self.experiment.methods.keys()]  # type: ignore

        self._complete_df = None
        self._base_df = None

    @staticmethod
    def get_result_selector(result_selector: Union[Experiment, "ResultsSelector"],
                            scenarios: Optional[list[str]] = None,
                            methods: Optional[list[str]] = None) -> "ResultsSelector":
        if isinstance(result_selector, Experiment):
            return ResultsSelector(result_selector, scenarios=scenarios, methods=methods)
        return result_selector

    def check_special_df(self, df: DataFrame):
        if not self.base_df.shape == df.shape:
            raise ValueError(f"Dataframe shape {df.shape} does not match base dataframe shape {self.base_df.shape}")
        if not list(self.base_df.dtypes) == list(df.dtypes):
            raise ValueError(f"Dataframe dtypes {df.dtypes} do not match base dataframe dtypes {self.base_df.dtypes}")

    @property
    def complete_df(self) -> DataFrame:
        if self._complete_df is None:
            data = [{"scenario": scenario} |
                    {k: v for k, v in self.experiment.get_scenario(scenario).result_tree.data.results.items()}
                    for scenario in self.experiment.scenario_aliases]

            self._complete_df = DataFrame(data)
        return self._complete_df

    @property
    def base_df(self) -> DataFrame:
        if self._base_df is None:
            data = [{"scenario": scenario} |
                    {k: v for k, v in self.experiment.get_scenario(scenario).result_tree.data.results.items() if
                     k in self.methods}
                    for scenario in self.scenarios]

            self._base_df = DataFrame(data)
        return self._base_df

    def normalized_df(self, normalize_with_all_scenarios: bool = True) -> DataFrame:
        used_df = self.complete_df if normalize_with_all_scenarios else self.base_df
        columns = used_df.columns[1:]
        normalized_df = DataFrame()
        normalized_df['scenario'] = used_df['scenario']
        for column in columns:
            normalized_df[column] = MinMaxScaler().fit_transform(used_df[column].to_numpy().reshape(-1, 1))

        if normalize_with_all_scenarios:
            # delete all rows where scenario is not in self.scenarios
            normalized_df = normalized_df[normalized_df['scenario'].isin(self.scenarios)]
        return normalized_df

    def method_label_names(self, short: bool = True, include_unit: bool = True) -> list[str]:
        exp_methods = self.experiment.methods
        return [((exp_methods[l].id[-1] if short else "\n".join(exp_methods)) +
                                                     ("\n" + exp_methods[l].bw_method.unit if include_unit else ""))
                for l in self.methods]

    def compare_to_baseline(self, baseline_data: ndarray):
        assert len(baseline_data) == len(self.methods)
        # Create a copy of the original dataframe, without modifying it in place
        baseline_df = self.base_df.copy()
        # Use loc to select the columns related to methods and divide them by baseline_data
        baseline_df[self.methods] = self.base_df[self.methods].div(baseline_data, axis=1)
        # Set the dtypes back to their original types
        original_dtypes = self.base_df[self.methods].dtypes
        for col, dtype in original_dtypes.items():
            # print(col, dtype)
            baseline_df[col] = baseline_df[col].astype(dtype)
        return baseline_df.reset_index(drop=True)

    def collect_tech_results(self, node_aliases: list[str]):
        df = DataFrame()
        for scenario in self.scenarios:
            scenario_results = self.experiment.get_scenario(scenario).result_tree
            for node_alias in node_aliases:
                node = scenario_results.find_subnode_by_name(node_alias)
                assert node is not None
                # add a row
                df = df._append({
                                    "scenario": scenario,
                                    "tech": node_alias,
                                } | {method: value for method, value in node.data.results.items() if
                                     method in self.methods},
                                ignore_index=True)

        return df



import json
from dataclasses import dataclass

from enbios2.analyse.const import indicator_map_file
from enbios2.generic.util import generate_levensthein_dict_map


@dataclass
class Indicator:
    name: str
    description: str
    unit: str
    abbr: str


def generate_indicator_map(indicator_names: list[str]) -> dict[str, dict]:
    indicators: list[dict[str, str]] = json.loads(indicator_map_file.read_text(encoding="utf-8"))
    map = generate_levensthein_dict_map(indicator_names, indicators, "name")
    # now create a map
    print(map)
    return map



from dataclasses import dataclass
from typing import Optional


@dataclass
class LCISpoldDefinition:
    Processor: str
    EcoinventFilename: str
    EcoinventCarrierName: Optional[str]

    def __post_init__(self):
        # when Filename does not end with .spold, add it
        if not self.EcoinventFilename.endswith(".spold"):
            self.EcoinventFilename += ".spold"


"""
Plot Enbios results
"""
import json
import logging
from csv import DictReader
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, Union, List, Dict, Tuple, Literal

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from pandas import DataFrame

from .const import scenario_col, processor_col, value_col, level_col, ABSOLUTE, RELATIVE, INDICATOR, PROCESSOR, \
    SCENARIO, \
    indicator_col, indicator_map_file
from .experiment_exporter import ExperimentExporter
from .old_util import generate_indicator_map


@dataclass
class ColumnMapper:
    scenario: str
    processor: str
    indicator: str
    value: str


class AnaExperiment:
    """
    Enbios experiment result-plotter class
    """

    def __init__(self, experiment_file: str,
                 column_mapper: ColumnMapper,
                 force_redo_split: bool = False,
                 auto_indicator_map: bool = True):
        """
        Initiate a ENBIOS experiment result analysis with a result file (duplicates should be removed)
        It will create a folder with the same name as the experiment-result file (and in the same folder)
        Upon first time, it will split the result file into scneario-based files and scenario-indicator based files.
        (which are used for plotting)
        :param experiment_file: (filtered) enbios result file
        :param force_redo_split: remove and redo splitting
        :param auto_indicator_map: try to create indicator map automatically using 
        """
        self.file_path = Path(experiment_file)
        self.column_mapper = column_mapper

        self.experiment_path = self.file_path.parent.joinpath(self.file_path.stem)

        if not self.file_path.exists():
            raise ValueError(f"File {self.file_path} does not exist")

        clean_file = self.file_path.parent.joinpath(self.file_path.stem + "_clean.csv")
        if clean_file.exists():
            self.file_path = clean_file
            self.use_clean = True

        self.complete_df = pd.read_csv(self.file_path, encoding="utf-8")
        self.exporter = ExperimentExporter(self)

        if auto_indicator_map:
            df_indicators: list[str] = list(self.complete_df[self.column_mapper.indicator].unique())
            self.indicator_map = generate_indicator_map(df_indicators)
        else:
            self.indicator_map: dict[str, dict] = {}

        # self.indicator_info = self.create_indicator_map(indicator_map_file)

        self.indicator_info: Dict[str, Dict[str, str]] = json.load(
            self.experiment_path.joinpath(indicator_map_file).open(encoding="utf-8"))

        # self.abbreviationReverseMap = {v["abbre"]: k for k, v in self.indicator_info.items()}

        # if force_redo_split:
        #     shutil.rmtree(self.experiment_path.joinpath("split-scenario"), ignore_errors=True)
        #     shutil.rmtree(self.experiment_path.joinpath("split-scenario-indicator"), ignore_errors=True)
        if self.setup_folder() or force_redo_split:
            self.remove_duplicates()
            # self._split_experiment()

    def col(self, col: Literal["scenario", "processor", "indicator", "value"]) -> str:
        return getattr(self.column_mapper, col)

    def setup_folder(self) -> bool:
        """
        creates all required directory.
        :return: true if the split folders were newly created
        """
        self.experiment_path.mkdir(exist_ok=True)
        self.experiment_path.joinpath("results").mkdir(exist_ok=True)
        newly_created_split_dirs = (not self.experiment_path.joinpath("split-scenario").exists()
                                    or not self.experiment_path.joinpath("split-scenario-indicator"))
        self.experiment_path.joinpath("split-scenario").mkdir(exist_ok=True)
        self.experiment_path.joinpath("split-scenario-indicator").mkdir(exist_ok=True)
        return newly_created_split_dirs

    def remove_duplicates(self):
        """
        Remove Duplicates, identifies by the Column Scenario, Processor, Indicator, Value

        :return:
        """
        print("Removing duplicates")
        df = self.get_complete_df()
        orig_len = len(df)
        df.drop_duplicates(subset=[scenario_col, processor_col, indicator_col, value_col], inplace=True)
        new_len = len(df)
        print(f"Removed {orig_len - new_len} duplicates from {orig_len} rows")
        new_name = self.file_path.stem + "_no_dupl.csv"
        df.to_csv(new_name, index=False)

    def create_indicator_map(self, indicator_map_path: Path) -> List[Dict[str, str]]:
        """
        create the info file, that contains , if possible abbreviation of the indicator
        :param indicator_map_path:
        :return:
        """
        complete_df = self.get_complete_df()
        # select all unique values in the indicator column
        indicators = complete_df[indicator_col].unique()
        result: Dict[str, Dict[str, str]] = {}
        for indicator in indicators:
            parts = indicator.split("_")
            if len(parts) > 1:
                abbre = parts[-1]
            else:
                print("No abbreviation found for indicator", indicator)
            result[indicator] = {"abbre": abbre}
        json.dump(result, indicator_map_path.open("w", encoding="utf-8"), indent=2)
        return result

    def _select_file(self, scenario: str, indicator: Optional[str] = None) -> Path:
        """
        select a Path (scenario) or (scenario-indicator), and throw an error if the file does not exists

        :param scenario: scenario-name
        :param indicator:  if indicator starts with "_" it should be the abbreviation of the indicator
        :return: Path of the file
        """
        if indicator:
            if indicator.startswith("_"):
                indicator_abbreviation = indicator[1:]
            else:
                indicator_abbreviation = self.indicator_info[indicator]["abbre"]
            path = self.experiment_path.joinpath("split-scenario-indicator", f"{scenario}_{indicator_abbreviation}.csv")
        else:
            path = self.experiment_path.joinpath("split-scenario", f"{scenario}.csv")
        if not path.exists():
            raise ValueError(f"File {path} does not exist")

        return path

    def get_data(self, df: DataFrame, filters: dict[str, any]) -> DataFrame:
        """

        :return: a list of dictionaries, with the columns as keys
        """
        if not filters:
            return df
        mask = True
        for column, value in filters.items():
            mask &= (df[column] == value)
        return df[mask]

    def get_complete_df(self):
        """
        get the complete dataframe of the experiment
        :return: pandas dataframe
        """
        if self.complete_df is None:
            data: List[Dict[str, any]] = list(DictReader(self.file_path.open(encoding="utf-8")))
            self.complete_df = pd.DataFrame(data)
        return self.complete_df

    def get_scenario_indicator_data(self,
                                    scenario: str,
                                    indicator: str,
                                    processor_name: Union[str, tuple, List]) -> pd.DataFrame:
        """
        Get a pandas dataframe for scenario-indicator-processor_name group.
        It will select all sub-processors and the next lower level.
        the resulting dataframe also includes a "relative" column with the values relative to the total impact
        :param scenario:
        :param indicator:
        :param processor_name: a string (sub-processes seperated by "." or a
            list/tuple of processors within the hierarchy
        :return: a pandas dataframe
        """
        data = self.get_data(scenario, indicator)
        complete_df = pd.DataFrame(data)
        # covert column "Value" to float
        complete_df[value_col] = complete_df[value_col].astype(float)
        # convert column "Dendrogram level" to int
        complete_df[level_col] = complete_df[level_col].astype(int)

        # split the df Processor column by "." and add the result into columns named
        # "level_x" where x is the level starting from 0.
        complete_df = complete_df.join(complete_df[processor_col].str.split(".", expand=True).add_prefix("level_"))
        # split the processor_name into a tuple by "."
        if isinstance(processor_name, str):
            processor_name = tuple(processor_name.split("."))

        # filter the df by the processor_name
        df = complete_df.copy()
        for i, name in enumerate(processor_name):
            df = df[df[f"level_{i}"] == name]

        # get the rows of the next level bytes filtering level_col column
        df = df[df[level_col] == len(processor_name)]
        # sum up the values the same level as the processor_name
        # others = complete_df[complete_df[level_col] == len(processor_name) - 2]
        # we don't need that. just get the total sum at the top level
        total_value = float(complete_df[complete_df[level_col] == 0][value_col])
        # print(total_value)
        # convert values in Value to float
        df[value_col] = df[value_col].astype(float)

        # throw out all columns but "level_x" and value_col
        df = df[[f"level_{len(processor_name)}", value_col]]
        # rename the "level_x" column to "Technology"
        df = df.rename(columns={f"level_{len(processor_name)}": "Technology"})
        # all a column relative to the total value
        df["relative"] = df[value_col] / total_value
        return df

    def plot_impacts(self,
                     scenarios: Union[str, List[str]],
                     indicator: str,
                     processor_name: Union[str, tuple, List],
                     _type: Literal["absolute", "relative"] = ABSOLUTE,
                     *,
                     yscale_max: float = None,
                     indicator_plot: Tuple[plt.Figure, plt.Axes, str] = None,
                     save: bool = True,
                     show: bool = False) -> Tuple[plt.Figure, plt.Axes, str]:
        """
        Create plots for a scenarios, indicator, processor_name group (multiple scnearios can be included)
        Plots absolute and/or relative values.
        Plots can be displayed and/or stored (<experiment_folder>/results/plots)
        :param scenarios: a scenario or a list of scenarios
        :param indicator: an indicator
        :param processor_name: a processor name or a tuple of processor names
        :param _type:
        :param indicator_plot:
        :param yscale_max:
        :param save: save the plots
        :param show: show the plots
        :return: a tuple of tuples of figure and axe
        """
        scenarios = scenarios if isinstance(scenarios, list) else [scenarios]

        scenarios_data = {
            scenario: self.get_scenario_indicator_data(scenario, indicator, processor_name)
            for scenario in scenarios
        }
        if indicator.startswith("_"):
            indicator_abbreviation = indicator[1:]
            indicator = self.abbreviationReverseMap[indicator[1:]]
        else:
            indicator_abbreviation = self.indicator_info[indicator]["abbre"]

        if indicator_plot is not None:
            _, _, passed_indicator_abbre = indicator_plot
            if passed_indicator_abbre != indicator_abbreviation:
                logging.warning("You are comparing different indicators!")

        # put them into one dataframe with scenario as index
        # df = pd.concat(scenarios_data.values(), keys=scenarios_data.keys())

        # create a second index from the Technology column
        # df = df.set_index("Technology", append=True)

        # TABLE VERSION
        # technologies = list(list(scenarios_data.values())[0]["Technology"])

        # get unique technologies
        technologies = scenarios_data[scenarios[0]]["Technology"].unique()

        # each scenario is a dataframe. I want to get the values of each technology
        weights = {
            tech: []
            for tech in technologies
        }
        for tech in technologies:
            # print(tech)
            for scenario, data in scenarios_data.items():
                # print(scenario)
                # print(data)
                selection = data[data["Technology"] == tech]
                if selection.empty:
                    logging.warning(f" No data for '{tech}' found in scenario '{scenario}'")
                    weights[tech].append(0)
                    continue
                if _type == ABSOLUTE:
                    weights[tech].append(selection[value_col].values[0])
                else:
                    weights[tech].append(selection["relative"].values[0])
        # plot them in bar-plots, for each scenario one bar, and for each technology stacked upon:
        width = 0.6

        fig, ax, _ = indicator_plot if indicator_plot else (*plt.subplots(), indicator)

        plt.xticks(rotation=-25)
        bottom = np.zeros(len(scenarios_data.keys()))

        for tech, weight_count in weights.items():
            _ = ax.bar(scenarios_data.keys(), weight_count, width, label=tech, bottom=bottom)
            bottom += weight_count
            if yscale_max:
                ax.set_ylim(0, yscale_max)

        indicator_data = self.indicator_info[indicator]
        ax.set_title(indicator_data["name"])
        ax.legend(loc="upper left", bbox_to_anchor=(1.01, 1))
        if _type == ABSOLUTE:
            indicator_unit = self.indicator_info[indicator]["unit"]
            plt.ylabel(indicator_unit)
        else:
            plt.ylabel("relative to total impact")

        plt.tight_layout()
        # save
        if save:
            self.experiment_path.joinpath("results", "plots").mkdir(exist_ok=True)
            plot_result_folder = self.experiment_path.joinpath("results", "plots")
            scenarios_name = self.exporter.scenario_name_generator(scenarios)
            processor_short_name = self.exporter.processor_name_generator(processor_name)
            base_name = f"{scenarios_name}__{indicator_abbreviation}__{processor_short_name}"
            filename = plot_result_folder.joinpath(f"{base_name}{'_REL' if _type == RELATIVE else ''}.png")
            fig.savefig(filename)

        if show:
            plt.show()

        return fig, ax, self.indicator_info[indicator]["abbre"]
        # return fig_abs, fig_rel

    def get_all_names(self, type_: Literal[INDICATOR, PROCESSOR, SCENARIO]) -> List[str]:

        def readExpFile(filename: str) -> List[str]:
            return json.load(self.experiment_path.joinpath(filename).open("r", encoding="utf-8"))

        if type_ == INDICATOR:
            return readExpFile("indicators.json")
        elif type_ == PROCESSOR:
            return readExpFile("processors.json")
        elif type_ == SCENARIO:
            return readExpFile("scenarios.json")



from huggingface_hub import login
from transformers import AutoModelForCausalLM, AutoTokenizer

from enbios2.settings import settings

from datasets import list_datasets, load_dataset
import datasets
# Print all the available datasets
# print(list_datasets())
# TODO WHERE TO?
# ds = load_dataset("ArmelR/stack-exchange-instruction") # home/.cache/hugging_face
ds = datasets.load_dataset("ArmelR/stack-exchange-instruction", split="train", streaming=True)

#
# login(settings.huggingface_token)
# checkpoint = "bigcode/starcoder"
# device = "cpu"  # for GPU usage or "cpu" for CPU usage
#
# tokenizer = AutoTokenizer.from_pretrained(checkpoint)
# # to save memory consider using fp16 or bf16 by specifying torch.dtype=torch.float16 for example
# model = AutoModelForCausalLM.from_pretrained(checkpoint).to(device)
#
# inputs = tokenizer.encode("def print_hello_world():", return_tensors="pt").to(device)
# outputs = model.generate(inputs)
# print(tokenizer.decode(outputs[0]))



from pint import UnitRegistry

from enbios2.init_appdir import get_init_appdir

ureg = UnitRegistry()

ecoinvent_units_file_path = get_init_appdir() / "ecoinvent_pint_unit_match.txt"

if not ecoinvent_units_file_path.exists():
    print(f"Creating 'ecoinvent_pint_unit_match' file at: {ecoinvent_units_file_path.as_posix()}")
    ecoinvent_units_file_path.touch()

ureg.load_definitions(ecoinvent_units_file_path)



from csv import DictReader

from flatten_dict import unflatten
from frictionless import Schema, Resource, validate, system
from frictionless.fields import NumberField, StringField

from enbios2.generic.files import ReadPath
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import (ExperimentDataIO, ExperimentData, ActivitiesDataRows,
                                              ExperimentMethodData, ActivitiesDataTypes, MethodsDataTypes)

activities_schema = Schema(fields=[
    StringField(name="alias"),
    StringField(name="database"),
    StringField(name="code"),
    StringField(name="name"),
    StringField(name="location"),
    StringField(name="unit"),
    NumberField(name="output.value", float_number=True),
    StringField(name="output.unit")
])

activities_unflatten_map = {
    "alias": "id.alias",
    "database": "id.database",
    "code": "id.code",
    "name": "id.name",
    "location": "id.location",
    "unit": "id.unit",
    "output.value": "output.value",
    "output.unit": "output.unit"
}

methods_schema = Schema(fields=[
    StringField(name="alias"),
    StringField(name="id"),
    StringField(name="id0"),
    StringField(name="id1"),
    StringField(name="id2"),
    StringField(name="id3")
])


# methods_fields = ["alias", "id", "id0", "id1", "id2", "id3"]


def unflatten_data(data: dict, structure_map: dict):
    res = {}
    for key, value in data.items():
        if key in structure_map:
            res[structure_map[key]] = value
        else:
            res[key] = value
    print(res)
    return unflatten(res, splitter="dot")


def parse_method_row(row: dict) -> ExperimentMethodData:
    id_: list[str] = []
    if row_id := row.get("id"):
        id_ = row_id.split(",")
    else:
        for i in range(4):
            if row_id := row.get(f"id{i}"):
                id_.append(row_id)
            else:
                break
    return ExperimentMethodData(alias=row.get("alias"), id=id_)


def read_experiment_io(data: dict) -> ExperimentData:
    exp_io = ExperimentDataIO(**data)
    experiment_data = ExperimentData(exp_io.bw_project)

    def get_abs_path(path: str) -> ReadPath:
        if exp_io.config.base_directory:
            return ReadPath(exp_io.config.base_directory) / path
        else:
            return ReadPath(path)

    if isinstance(exp_io.activities, str):
        activities_file: ReadPath = get_abs_path(exp_io.activities)
        if activities_file.suffix == ".json":
            data = activities_file.read_data()
            experiment_data.activities = data
        elif activities_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(path=activities_file.as_posix(), schema=activities_schema)
                report = validate(resource)
                if not report.valid:
                    raise Exception(f"Invalid activities file: {exp_io.activities}; errors: {report.task.errors}")
                experiment_data.activities = ActivitiesDataRows(
                    list((unflatten_data(r, activities_unflatten_map) for r in resource.read_rows())))
    else:
        experiment_data.activities = ActivitiesDataTypes(exp_io.activities)

    if isinstance(exp_io.methods, str):
        methods_file: ReadPath = get_abs_path(exp_io.methods)
        if methods_file.suffix == ".json":
            data = methods_file.read_data()
            experiment_data.methods = data
        if methods_file.suffix == ".csv":
            with system.use_context(trusted=True):
                resource: Resource = Resource(path=methods_file.as_posix(), schema=methods_schema)
                report = validate(resource)
                if not report.valid:
                    raise Exception(f"Invalid methods file: {exp_io.methods}; errors: {report.task.errors}")
                experiment_data.methods = list((parse_method_row(r) for r in resource.read_rows()))
    else:
        experiment_data.methods = MethodsDataTypes(exp_io.methods)

    if isinstance(exp_io.hierarchy, str):
        hierarchy_file: ReadPath = get_abs_path(exp_io.hierarchy)
        if hierarchy_file.suffix == ".json":
            data = hierarchy_file.read_data()
            experiment_data.hierarchy = data
        else:
            raise Exception(f"Invalid hierarchy file: {exp_io.hierarchy}")
    return experiment_data


if __name__ == "__main__":
    scenario_data = {
        "bw_project": "uab_bw_ei39",
        "activities_config": {
            "default_database": "ei391"
        },
        "config": {
            "base_directory": "/mnt/SSD/projects/LIVENLab/enbios2/data/test_data/experiment_separated/a/"
        },
        "activities": "single_activity.csv",
        "methods": "methods.csv",
        "hierarchy": {
            "energy": [
                "single_activity"
            ]
        }
    }
    exp: ExperimentData = read_experiment_io(scenario_data)

    reader = DictReader(open("/mnt/SSD/projects/LIVENLab/enbios2/data/templates/experiment_activity_data.csv"))

    node = BasicTreeNode.from_compact_dict(scenario_data["hierarchy"])



from pathlib import Path

import appdirs


def get_init_appdir() -> Path:
    enbios_appdir = Path(appdirs.user_data_dir("enbios2"))
    if not enbios_appdir.exists():
        enbios_appdir.mkdir(parents=True, exist_ok=True)
        print(f"Creating 'enbios2' appdir at: {enbios_appdir.as_posix()}")
    return enbios_appdir




import json
from csv import DictWriter
from typing import Dict, Union, List, TYPE_CHECKING

from pandas import DataFrame

# import Experiment but just for type hinting
if TYPE_CHECKING:
    from enbios2.analyse.ana_experiment import AnaExperiment

# from const import processor_col, value_col


class ExperimentExporter:

    def __init__(self, experiment: AnaExperiment):
        self.experiment = experiment

    # def build_simple_tree(self, scenario: str, indicator: str, save: bool = True) -> Dict[str, any]:
    #     """
    #     builds a simple tree with the structure of each node:
    #     {name: full-process-name, children: {short-name: <child-node>}
    #     :param scenario:
    #     :param indicator:
    #     :param save:
    #     :return:
    #     """
    #     data = self.experiment.get_data(scenario, indicator)
    #     tree = {}
    #     for row in data:
    #         # get the value in column called processor_col
    #         proc = row[processor_col]
    #         proc_tuple = tuple(proc.split("."))
    #         if len(proc_tuple) == 1:
    #             tree[proc] = {"name": proc, "children": {}, "value": float(row[value_col])}
    #         else:
    #             # find all parent nodes starting from the root
    #             parent = tree
    #             for i in range(len(proc_tuple) - 1):
    #                 parent = parent[proc_tuple[i]]["children"]
    #             parent[proc_tuple[-1]] = {"name": proc, "children": {}, "value": float(row[value_col])}
    #
    #     if save:
    #         self.experiment.experiment_path.joinpath("results", "simple_trees").mkdir(exist_ok=True)
    #         indicator_name = self.experiment.indicator_info[indicator]["abbre"]
    #         file_path = self.experiment.experiment_path.joinpath("results",
    #                                                              "simple_trees",
    #                                                              f"{scenario}_{indicator_name}.json")
    #         json.dump(tree, file_path.open("w", encoding="utf-8"), indent=2, ensure_ascii=False)
    #     return tree

    def d3sanki(self, scenario: str, indicator: str, save: bool = True) -> List[Dict[str, Union[str, float]]]:
        """
        creates a csv file with the following columns:
        source	target	value

        :param scenario:
        :param indicator:
        :param save: save to file
        :return: result data
        """
        col_sce = self.experiment.col("scenario")
        col_ind = self.experiment.col("indicator")
        col_pro = self.experiment.col("processor")
        col_val = self.experiment.col("value")
        data: DataFrame = self.experiment.get_data(self.experiment.complete_df, {col_sce: scenario, col_ind: indicator})

        result_data = []
        for row in data.iterrows():
            row = row[1].to_dict()
            proc = row[col_pro]
            proc_tuple = proc.split(".")
            if len(proc_tuple) > 1:
                result_node = {
                    "source": proc_tuple[-2],
                    "target": proc_tuple[-1],
                    "target_level": len(proc_tuple) - 1,
                    "value": float(row[col_val])
                }
                result_data.append(result_node)
        # print(json.dumps(result_data, indent=2))
        if save:
            dir = self.experiment.experiment_path.joinpath("results", "d3sanki")
            dir.mkdir(exist_ok=True)

            # indicator_abbre = indicator[1:] if indicator.startswith("_") else self.experiment.indicator_info[indicator][
            #     "abbre"]
            if self.experiment.indicator_map:
                indicator_abbre = self.experiment.indicator_map[indicator]["abbre"]
            else:
                indicator_abbre = indicator
            writer = DictWriter(dir.joinpath(f"{scenario}_{indicator_abbre}.csv").open("w", encoding="utf-8"),
                                fieldnames=["source", "target", "value", "target_level"])
            writer.writeheader()
            writer.writerows(result_data)
        return result_data

    def scenario_name_generator(self, scenarios: List[str]):
        """
        This is some very specific code, for scenarios in the decades after 2000
        :param scenarios:
        :return:
        """
        # scenario_splits = [s.split("_") for s in scenarios]
        # scenario_types = set([s[0] for s in scenario_splits])
        sce_grouped = {}
        for sce in scenarios:
            sce_type, sce_year = sce.split("_")
            sce_grouped.setdefault(sce_type, []).append(sce_year[2:])
        # create a string for each scenario type
        sce_strs = []
        for sce_type, sce_years in sce_grouped.items():
            sce_strs.append(f"{sce_type[:5]}_{'_'.join(sce_years)}")
        return "_".join(sce_strs)

    def processor_name_generator(self, processor: str):
        proc_parts = processor.split(".")
        if len(proc_parts) == 1:
            return processor
        return "_".join(proc_parts[1:])



import json
from typing import Tuple, Optional

import bw2data
import networkx
import numpy
import orjson as orjson
from bw2data.backends import ActivityDataset, ExchangeDataset, Activity
from networkx import DiGraph
from playhouse.shortcuts import model_to_dict

# print(projects)
# projects.set_current("ecoi_dbs")
# db = Database("cutoff391")

from enbios2.bw2.project_index import print_bw_index, set_bw_current_project
from enbios2.const import BASE_DATA_PATH


def check_unique_codes():
    activity_codes = [a.code for a in ActivityDataset.select(ActivityDataset.code)]
    assert len(activity_codes) == len(set(activity_codes))


def get_tree(code: str,
             keep_exchange_type: list[str] = None,
             check_unique_code: bool = True,
             max_level: int = -1) -> Tuple[set[str], list[int]]:
    """
    Get all nodes that are connected to the given node. (as inputs)
    :param code: root code
    :param keep_exchange_type: keep the exchanges of the given type
    :param check_unique_code: check if the codes are unique
    :param max_level: max level to go down. -1 for infinite
    :return:
    """
    if check_unique_code:
        check_unique_codes()

    # all visited nodes (codes)
    visited: set[str] = set()
    # nodes to visit next
    to_visit = {code}
    # nodes to visit in the next iteration
    to_visit_next = set()
    # all exchanges (eventually filtered by type). just ids for memory efficiency
    all_exchanges: list[int] = []
    current_level: int = 0

    while len(to_visit) and max_level != 0:
        # get all exchanges that we could currently reach
        exchanges = list(
            ExchangeDataset.select(ExchangeDataset.id, ExchangeDataset.input_code, ExchangeDataset.output_code).where(
                ExchangeDataset.output_code.in_(to_visit)))
        # add nodes that we did not visit yet for the next iteration
        for exc in exchanges:
            if exc.input_code == exc.output_code:
                continue
            if exc.input_code not in visited:
                to_visit_next.add(exc.input_code)
        # updated visited nodes
        visited.update(to_visit)
        # save exchanges
        if not keep_exchange_type:
            all_exchanges.extend([e.id for e in exchanges])
        else:
            all_exchanges.extend([exc.id for exc in exchanges if exc.type in keep_exchange_type])
        # update nodes to visit for next iteration
        to_visit = to_visit_next.copy()
        # reset nodes to visit in next iteration
        to_visit_next.clear()

        print(f"visited: {len(visited)}, total exchanges: {len(all_exchanges)}, next nodes: {len(to_visit)}")
        max_level -= 1
        current_level += 1
    return visited, all_exchanges


def get_tree_with_levels(code: str,
                         keep_exchange_type: Optional[list[str]] = None,
                         check_unique_code: Optional[bool] = True,
                         max_level: Optional[int] = -1) -> Tuple[dict[str, int], dict[str, set[str]], list[int]]:
    """
    Get all nodes that are connected to the given node. (as inputs)
    :param code: root code
    :param keep_exchange_type: keep the exchanges of the given type
    :param check_unique_code: check if the codes are unique
    :param max_level: max level to go down. -1 for infinite
    :return:
    """
    if check_unique_code:
        check_unique_codes()

    # all visited nodes (codes)
    visited: set[str] = set[str]()
    # keep a dict of code -> level
    level: dict[str, int] = dict()
    # keep a dict. code -> all input codes
    inputs: dict[str, set[str]] = dict()
    # nodes to visit next
    to_visit = {code}
    # nodes to visit in the next iteration
    # all exchanges (eventually filtered by type). just ids for memory efficiency
    all_exchanges: list[int] = []
    current_level: int = 0

    while len(to_visit) and max_level != 0:
        print("---")
        # get all exchanges that we could currently reach
        visited.update(to_visit)
        for v in to_visit:
            level[v] = current_level
            inputs[v] = set()

        exchanges = list(
            ExchangeDataset.select(ExchangeDataset.id,
                                   ExchangeDataset.input_code,
                                   ExchangeDataset.output_code,
                                   ExchangeDataset.type).where(
                ExchangeDataset.output_code.in_(to_visit)))

        to_visit.clear()
        # add nodes that we did not visit yet for the next iteration
        for exc in exchanges:
            # ignore if production
            if exc.input_code == exc.output_code:
                continue
            if exc.input_code not in visited:
                to_visit.add(exc.input_code)
                inputs[exc.output_code].add(exc.input_code)
            else:
                # get level of visited node
                def rec_inc(node: str, new_level):
                    level[node] = new_level
                    for inp in inputs.get(node, []):
                        rec_inc(inp, new_level + 1)

                rec_inc(exc.input_code, level[exc.output_code] + 1)
        # save exchanges
        if not keep_exchange_type:
            all_exchanges.extend([e.id for e in exchanges])
        else:
            all_exchanges.extend([exc.id for exc in exchanges if exc.type in keep_exchange_type])
        print(f"visited: {len(visited)}, total exchanges: {len(all_exchanges)}, next nodes: {len(to_visit)}")
        max_level -= 1
        current_level += 1
    # sort by the second element of the tuple (level)
    return level, inputs, all_exchanges


if __name__ == "__main__":
    print_bw_index()
    set_bw_current_project("cutoff", "3.9.1")
    print(list(bw2data.databases))
    # random_act = db.random()
    db = bw2data.Database("cutoff_3.9.1_default")
    random_act: Activity = db.get_node("46875148bb5fbac9cad7452d020a80de")
    print(random_act)
    # # print(info_exchanges(random_act))
    #
    calc = True
    # 'non-ionic surfactant production, ethylene oxide derivate' (kilogram, GLO, None)

    if calc:
        levels, inputs, exchanges = get_tree_with_levels(random_act["code"])
        # json.dump((list(visited_with_levels), list(exchanges)), open("temp.json", "w"))
    else:
        visited_with_levels, exchanges = json.load(open("temp.json", "r"))


    # visited, levels = zip(*visited_with_levels)
    # activity_iter = iter_activities_by_codes(iter(visited))

    # graph = grap_nodes(activity_iter, iter_exchange_by_ids(iter(exchanges)), level_infos=levels)
    # json.dumps(graph)

    def default(obj):
        if isinstance(obj, set):
            return list(obj)
        if isinstance(obj, numpy.ndarray):
            return list(obj)
        raise TypeError


    (BASE_DATA_PATH / "fast_tree_data.json").write_text(str(orjson.dumps({
        "levels": levels,
        "inputs": inputs,
        "exchanges": exchanges
    }, default=default)), encoding="utf-8")

    activities = list(ActivityDataset.select().where(ActivityDataset.code.in_(list((levels.keys())))))
    (BASE_DATA_PATH / "fast_tree_activities_data.json").write_text(str(orjson.dumps(
        [[a.key, a.id, a.name, a.location, a.product, a.type, a.data.get("unit")] for a in activities],
        default=default)), encoding="utf-8")

    uuid2id = {a.code: a.id for a in activities}

    (BASE_DATA_PATH / "fast_tree_activities_data_keyed.json").write_text(str(orjson.dumps(
        {str(a.id): [a.database, a.id, a.name, a.location, a.product, a.type, a.data.get("unit"),
                     [uuid2id[i] for i in inputs[a.code]]] for a in activities},
        default=default)), encoding="utf-8")

    graph = DiGraph()
    graph.add_nodes_from([(act.id, {"name": act.name}) for act in activities])
    for act in activities:
        for i in inputs[act.code]:
            graph.add_edge(uuid2id[i], act.id)

    # this will take a long time...
    pos = networkx.spring_layout(graph)
    (BASE_DATA_PATH / "fast_tree_graph_positions.json").write_text(str(orjson.dumps({str(k): v for k, v in pos.items()},
                                                                                    default=default)), encoding="utf-8")



# import tempfile
from pathlib import Path

import pandas as pd
from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy

from sklearn.cluster import KMeans

import matplotlib.pyplot as plt
from sklearn.preprocessing import MinMaxScaler
from matplotlib.colors import ListedColormap


def clean_data(df: DataFrame) -> DataFrame:
    # kick out scope = Internal
    df = df[df.Scope != "Internal"]
    # kick out scope, period, unit cuz it is all the same
    df.drop(columns=["Scope", "Period", "Unit"], inplace=True)
    # kick out all where Processor is "environment"
    df = df[df.Processor != "environment"]
    # transform the Scenario column to int
    df["Scenario"] = df['Scenario'].str.replace('_', '').astype(int)
    # Optional, calculate the level, and System group.
    # add a new column of type int, by splitting the processor by "." and counting the length of the array
    df["Level"] = df["Processor"].str.split(".").apply(len).astype(int)

    return df


def add_systems_to_total_rows(df: DataFrame,
                              total_system_name: str,
                              group: list[str] = ("Scenario", "Processor", "Indicator"),
                              ) -> DataFrame:
    group = list(group)  # cuz pandas doesnt like tuples
    df["Systemgroup"] = "region"
    # add a row for the total of the scenario
    grouped_df = df.groupby(group).agg({"Value": "sum"}).reset_index()
    grouped_df["System"] = total_system_name
    grouped_df["Systemgroup"] = "total"
    grouped_df["Level"] = grouped_df["Processor"].str.split(".").apply(len).astype(int)

    # Reorder the columns to match the original DataFrame
    grouped_df = grouped_df[["Scenario", "System", "Systemgroup", "Processor", "Level", "Indicator", "Value"]]
    # Concatenate the original DataFrame with the aggregated data
    result_df = pd.concat([df, grouped_df]).reset_index(drop=True)
    result_df = result_df.sort_values(group)
    return result_df


def add_normalized_col(df: DataFrame, groups: list[str] = ("Level", "Indicator")) -> DataFrame:
    def apply_min_max_scaler(group: DataFrameGroupBy) -> DataFrameGroupBy:
        scaler = MinMaxScaler()
        scaled_values = scaler.fit_transform(group['Value'].to_numpy().reshape(-1, 1))
        group['norm_value'] = scaled_values
        return group

    return df.groupby(list(groups)).apply(apply_min_max_scaler).reset_index(drop=True)


def cluster_on_processor(df: DataFrame, processor: str, n_clusters=3) -> DataFrame:
    if "norm_value" not in df.columns:
        raise ValueError("The DataFrame must contain a column named 'norm_value', run 'add_normalized_col' first.")
    # Filter the DataFrame based on the indicator and processor
    df_filtered = df[df['Processor'] == processor]
    df_pivot = df_filtered.pivot_table(index='Scenario', columns='Indicator', values='norm_value')

    # Run KMeans clustering
    X = df_pivot.values
    kmeans = KMeans(n_clusters=n_clusters, n_init=10)
    cluster_labels = kmeans.fit_predict(X)

    # Create a new DataFrame with the required format
    df_result = df_pivot.reset_index()
    df_result.columns = [f"Value_{col}" if col != 'Scenario' else col for col in df_result.columns]
    df_result['Label'] = cluster_labels
    df_result['Processor'] = processor

    # Merge the original values back into the DataFrame
    df_result = df_result.melt(id_vars=['Scenario', 'Label', 'Processor'], var_name='Indicator',
                               value_name='scaled_value')
    df_result['Indicator'] = df_result['Indicator'].str.replace('Value_', '')
    df_result = df_result.merge(df_filtered[['Scenario', 'Indicator', 'Value']], on=['Scenario', 'Indicator'],
                                how='left')

    return df_result


def plot_heatmaps_stacked(df_clustered: DataFrame,
                          plot_image_path: str,
                          sort: bool = True):

    try:
        import seaborn as sns
    except ImportError:
        raise ImportError("The seaborn package is required to run this function.")

    if not (parent_path := Path(plot_image_path).parent).exists():
        raise ValueError(f"The directory {parent_path.as_posix()} does not exist.")
    indicators = df_clustered['Indicator'].unique()
    n_indicators = len(indicators)
    scenarios = df_clustered['Scenario'].unique()
    n_scenarios = len(scenarios)

    # Create a colormap based on the cluster labels
    unique_labels = df_clustered['Label'].unique()
    colormap = ListedColormap(sns.color_palette("husl", len(unique_labels)))

    # Set the style and size of the plots
    sns.set(style='white')
    fig, axes = plt.subplots(n_indicators, 1, figsize=(n_scenarios / 6, n_indicators * 3), sharex=True)

    # Create a heatmap for each indicator
    for idx, indicator in enumerate(indicators):
        # Sort the DataFrame by cluster label for the current indicator
        df_indicator = df_clustered[df_clustered['Indicator'] == indicator].sort_values(by='Label')

        # Sort the scenarios based on the label/group
        if sort:
            sorted_scenarios = df_indicator.drop_duplicates(subset='Scenario', keep='first').sort_values(by='Label')[
                'Scenario'].tolist()
        else:
            sorted_scenarios = df_indicator.drop_duplicates(subset='Scenario', keep='first')['Scenario'].tolist()

        data = df_indicator.pivot_table(index='Indicator', columns='Scenario', values='Value')[sorted_scenarios]

        # Create a custom color mapping for the x-axis labels based on the cluster labels
        cluster_colors = [colormap(
            df_clustered.loc[(df_clustered['Scenario'] == s) & (df_clustered['Indicator'] == indicator), 'Label'].iloc[
                0]) for s in sorted_scenarios]

        sns.heatmap(data, cmap='viridis', ax=axes[idx], cbar_kws={'label': 'Value'})

        # Set plot title and labels
        axes[idx].set_title(f'Heatmap of {indicator} by Scenario')
        axes[idx].set_xticks(range(len(sorted_scenarios)))
        axes[idx].set_xticklabels(sorted_scenarios, fontsize=8, ha="left")

        # Set custom colors for each xticklabel
        for tick, color in zip(axes[idx].get_xticklabels(), cluster_colors):
            tick.set_color(color)
            # tick.set_ha("center")
        if idx == 0:
            axes[idx].set_xlabel('Scenario')

        axes[idx].set_ylabel('')

    # Show the plot
    plt.tight_layout()
    # plt.show()
    fig.savefig(plot_image_path)
    return plt



from enbios2.const import BASE_DATA_PATH

scenario_col = "Scenario"
value_col = "Value"
level_col = "Dendrogram level"
processor_col = "Processor"
indicator_col = "Indicator"

ABSOLUTE = "absolute"
RELATIVE = "relative"


INDICATOR = "indicator"
PROCESSOR = "processor"
SCENARIO = "scenario"

# the filee which includes the information about the indicators
indicator_map_file = BASE_DATA_PATH / "generic/indicator_info.json"



from typing import Optional

from pydantic import BaseSettings


class Settings(BaseSettings):
    huggingface_token: Optional[str] = None

    class Config:
        # `.env.prod` takes priority over `.env`
        env_file = '.env', 'template.env'


settings = Settings()



import math
import time
from dataclasses import dataclass, field
from datetime import timedelta
from typing import Optional, Union, TYPE_CHECKING, Any

from bw2data.backends import Activity
from numpy import ndarray
from pint import DimensionalityError, Quantity, UndefinedUnitError
from pint.facets.plain import PlainQuantity

from enbios2.base.stacked_MultiLCA import StackedMultiLCA
from enbios2.base.unit_registry import ureg
from enbios2.bw2.util import bw_unit_fix
from enbios2.generic.enbios2_logging import get_logger
from enbios2.generic.files import PathLike

# for type hinting
if TYPE_CHECKING:
    from enbios2.base.experiment import Experiment
from enbios2.generic.tree.basic_tree import BasicTreeNode
from enbios2.models.experiment_models import (BWCalculationSetup,
                                              ScenarioResultNodeData, ExperimentMethodPrepData, Activity_Outputs)

logger = get_logger(__file__)


@dataclass
class Scenario:
    experiment: "Experiment"
    alias: str
    result_tree: BasicTreeNode[ScenarioResultNodeData]
    results: Optional[ndarray] = None
    _has_run: bool = False
    # this should be a simpler type - just str: float
    activities_outputs: Activity_Outputs = field(default_factory=dict)
    methods: Optional[dict[str, ExperimentMethodPrepData]] = None
    _execution_time: float = float('NaN')

    def prepare_tree(self):
        # activity_nodes = list(self.result_tree.get_leaves())
        activities_simple_ids = list(self.activities_outputs.keys())
        for result_index, simple_id in enumerate(activities_simple_ids):
            alias = simple_id.alias
            bw_activity = self.experiment.get_activity(alias).bw_activity
            try:
                activity_node = self.result_tree.find_subnode_by_name(alias)
            except StopIteration:
                raise ValueError(f"Activity {alias} not found in result tree")
            activity_node._data = ScenarioResultNodeData(output=(bw_unit_fix(bw_activity['unit']),
                                                                 self.activities_outputs[simple_id]))
            if self.experiment.config.include_bw_activity_in_nodes:
                activity_node._data.bw_activity = bw_activity
        self.result_tree.recursive_apply(Scenario._recursive_resolve_outputs, depth_first=True)

    def _create_bw_calculation_setup(self, register: bool = True) -> BWCalculationSetup:
        inventory: list[dict[Activity, float]] = []
        for activity_alias, act_out in self.activities_outputs.items():
            bw_activity = self.experiment.get_activity(activity_alias.alias).bw_activity
            inventory.append({bw_activity: act_out})

        methods = [m.id for m in self._get_methods().values()]
        calculation_setup = BWCalculationSetup(self.alias, inventory, methods)
        if register:
            calculation_setup.register()
        return calculation_setup

    def create_results_to_technology_tree(self, results: ndarray) -> BasicTreeNode[ScenarioResultNodeData]:
        """
        Add results to the technology tree, for each method
        """

        def recursive_resolve_node(node: BasicTreeNode[ScenarioResultNodeData], _: Any = None):
            for child in node.children:
                if child._data:
                    for key, value in child._data.results.items():
                        assert node._data
                        if node._data.results.get(key) is None:
                            node._data.results[key] = 0
                        node._data.results[key] += value

        if not self.result_tree:
            raise ValueError(f"Scenario '{self.alias}' has no results...")
        activities_simple_ids = list(self.activities_outputs.keys())

        methods_aliases: list[str] = list(self._get_methods().keys())
        for result_index, simple_id in enumerate(activities_simple_ids):
            alias = simple_id.alias
            activity_node = self.result_tree.find_subnode_by_name(alias)
            assert activity_node
            # bw_activity = self.experiment.get_activity(alias).bw_activity
            # activity_node = next(
            #     filter(lambda node: node.temp_data()["activity"].bw_activity == bw_activity, activity_nodes))
            for method_index, method in enumerate(methods_aliases):
                activity_node._data.results[method] = results[result_index][method_index]

        self.result_tree.recursive_apply(recursive_resolve_node, depth_first=True)
        return self.result_tree

    def _get_methods(self) -> dict[str, ExperimentMethodPrepData]:
        if self.methods:
            return self.methods
        else:
            return self.experiment.methods

    @staticmethod
    def _recursive_resolve_outputs(node: BasicTreeNode[ScenarioResultNodeData], _: Optional[Any] = None):
        # todo, does this takes default values when an activity is not defined in the scenario?
        if node.is_leaf:
            return
        node_output: Optional[Union[Quantity, PlainQuantity]] = None
        for child in node.children:
            # todo, this should be removable
            if not child._data:
                raise ValueError(f"Node {child.name} has no data")
            activity_output = child.data.output[0]
            if activity_output is None:
                node_output = None
                logger.warning(
                    f"No output unit of activity '{child.name}'.")
                break
            output = None
            try:
                output = ureg.parse_expression(activity_output) * child.data.output[1]
                if not node_output:
                    node_output = output
                else:
                    node_output += output
            except UndefinedUnitError as err:
                logger.error(
                    f"Cannot parse output unit '{activity_output}' of activity {child.name}. {err}. "
                    f"Consider the unit definition to 'enbios2/base/unit_registry.py'")
                node_output = None
                break
            except DimensionalityError as err:
                set_base_unit = node_output.to_base_units() if node_output else ""
                logger.warning(f"Cannot aggregate output to parent: {node.name}. "
                               f"From earlier children the base unit is {set_base_unit} "
                               f"and from {child.name} it is {output}."
                               f" {err}")
                node_output = None
                break
        if node_output:
            node_output = node_output.to_compact()
            node.set_data(ScenarioResultNodeData(output=(str(node_output.units), node_output.magnitude)))
        else:
            node.set_data(ScenarioResultNodeData())
            logger.warning(f"No output for node: {node.name}")

    def run(self) -> BasicTreeNode[ScenarioResultNodeData]:
        if not self._get_methods():
            raise ValueError(f"Scenario '{self.alias}' has no methods")
        logger.info(f"Running scenario '{self.alias}'")
        start_time = time.time()
        bw_calc_setup = self._create_bw_calculation_setup()
        results: ndarray = StackedMultiLCA(bw_calc_setup).results
        result_tree = self.set_results(results)
        self._execution_time = time.time() - start_time
        return result_tree

    @property
    def execution_time(self) -> str:
        if not math.isnan(self._execution_time):
            return str(timedelta(seconds=int(self._execution_time)))
        else:
            return "not run"

    def reset_execution_time(self):
        self._execution_time = float('NaN')

    def set_results(self, results: ndarray) -> BasicTreeNode[ScenarioResultNodeData]:
        if self.experiment.config.store_raw_results:
            self.results = results
        self.result_tree = self.create_results_to_technology_tree(results)
        self._has_run = True
        return self.result_tree

    def wrapper_data_serializer(self, include_method_units: bool = False):

        method_alias2units: dict[str, str] = {
            method_alias: method_info.bw_method.unit
            for method_alias, method_info in self.experiment.methods.items()
        }

        def data_serializer(data: ScenarioResultNodeData) -> dict:
            result: dict[str, Union[str, float]] = {}
            if data.output:
                result["unit"] = data.output[0] or ""
                result['amount'] = data.output[1] or ""
            if not include_method_units:
                return result | data.results
            else:
                for method_alias, value in data.results.items():
                    final_name = f"{method_alias} ({method_alias2units[str(method_alias)]})"
                    result[final_name] = value
                return result

        return data_serializer

    def results_to_csv(self,
                       file_path: PathLike,
                       level_names: Optional[list[str]] = None,
                       include_method_units: bool = False,
                       warn_no_results: bool = True):
        """
        Save the results (as tree) to a csv file
         :param file_path:  path to save the results to
         :param level_names: names of the levels to include in the csv (must not match length of levels)
         :param include_method_units:
        """
        if not self.result_tree:
            raise ValueError(f"Scenario '{self.alias}' has no results")

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.alias}' has not been run yet")

        self.result_tree.to_csv(file_path,
                                include_data=True,
                                level_names=level_names,
                                data_serializer=self.wrapper_data_serializer(include_method_units))

    def result_to_dict(self, include_output: bool = True, warn_no_results: bool = True) -> dict[str, Any]:

        def data_serializer(data: ScenarioResultNodeData) -> dict:
            result: dict[str, Any] = {
                "results": data.results
            }
            if include_output:
                result["output"] = {
                    "unit": data.output[0],
                    'amount': data.output[1]
                }
            if data.bw_activity:
                result["bw_activity"] = data.bw_activity["code"]

            return result

        def recursive_transform(node: BasicTreeNode[ScenarioResultNodeData]) -> dict:
            result: dict[str, Any] = {"alias": node.name,
                                      **data_serializer(node._data)}
            if node.children:
                result["children"] = [recursive_transform(child) for child in node.children]
            return result

        if warn_no_results and not self._has_run:
            logger.warning(f"Scenario '{self.alias}' has not been run yet")
        return recursive_transform(self.result_tree.copy())



from copy import copy
from dataclasses import asdict, dataclass, field
from typing import Optional, Union

import bw2data
from bw2data.backends import Activity
from pydantic import Field
from pydantic.dataclasses import dataclass as pydantic_dataclass

from enbios2.bw2.util import get_activity


@pydantic_dataclass
class EcoInventSimpleIndex:
    version: str
    system_model: str


class StrictInputConfig:
    validate_assignment = True
    extra = "forbid"


class OperationConfig:
    validate_assignment = True
    arbitrary_types_allowed = True


@pydantic_dataclass(config=StrictInputConfig)
class ActivityOutput:
    unit: str
    magnitude: float = 1.0


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentActivityId:
    database: Optional[str] = None
    code: Optional[str] = None
    # search and filter
    name: Optional[str] = None
    location: Optional[str] = None
    # additional filter
    unit: Optional[str] = None
    # internal-name
    alias: Optional[str] = None

    def get_bw_activity(self, allow_multiple: bool = False) -> Union[Activity, list[Activity]]:
        if self.code:
            if not self.database:
                return get_activity(self.code)
            else:
                return bw2data.Database(self.database).get(self.code)
        elif self.name:
            filters = {}
            if self.location:
                filters["location"] = self.location
                assert self.database in bw2data.databases, f"database {self.database} not found"
                search_results = bw2data.Database(self.database).search(self.name, filter=filters)
            else:
                search_results = bw2data.Database(self.database).search(self.name)
            if self.unit:
                search_results = list(filter(lambda a: a["unit"] == self.unit, search_results))
            assert len(search_results) == 0, (f"No results for brightway activity-search:"
                                              f" {(self.name, self.location, self.unit)}")
            if len(search_results) > 1:
                if allow_multiple:
                    return search_results
                assert False, (f"results : {len(search_results)} for brightway activity-search:"
                               f" {(self.name, self.location, self.unit)}. Results are: {search_results}")
            return search_results[0]
        else:
            raise ValueError("No code or name specified")

    def fill_empty_fields(self, fields: Optional[list[Union[str, tuple[str, str]]]] = None, **kwargs):
        if not fields:
            fields = []
        for field in fields:
            if isinstance(field, tuple):
                if not getattr(self, field[0]):
                    setattr(self, field[0], kwargs[field[1]])
            else:
                if not getattr(self, field):
                    setattr(self, field, kwargs[field])


@pydantic_dataclass(config=StrictInputConfig)
class SimpleScenarioActivityId:
    name: str
    code: str
    alias: str

    def __hash__(self):
        return hash(self.code)


# this is just for the schema to accept an array.
ExperimentActivityOutputArray = tuple[str, float]

ExperimentActivityOutput = Union[ActivityOutput, ExperimentActivityOutputArray]


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentActivityData:
    """
    This is the dataclass for the activities in the experiment.
    the id, is
    """
    id: ExperimentActivityId = Field(..., description="The identifies (method to find) an activity")
    output: Optional[ExperimentActivityOutput] = Field(None, description="The default output of the activity")
    orig_id: Optional[ExperimentActivityId] = Field(None, description="Temporary copy of the id")

    @property
    def alias(self):
        return self.id.alias

    def check_exist(self, default_id_attr: Optional[ExperimentActivityId] = None,
                    required_output: bool = False) -> "ExtendedExperimentActivityData":
        """
        This method checks if the activity exists in the database by several ways.
        :param default_id_attr:
        :param required_output:
        :return:
        """
        orig_id: ExperimentActivityId = copy(self.id)
        bw_activity: Optional[Activity] = None

        if not self.id.database:
            if default_id_attr:
                self.id.database = default_id_attr.database
        # assert result.id.database in bd.databases,
        # f"activity database does not exist: '{self.id.database}' for {self.id}"
        # todo, is the needed?
        default_dict = asdict(default_id_attr) if default_id_attr else {}
        self.id.fill_empty_fields(["alias"], **default_dict)
        if self.id.code:
            if self.id.database:
                bw_activity = bw2data.Database(self.id.database).get_node(self.id.code)
            else:
                bw_activity = get_activity(self.id.code)
        elif self.id.name:
            # assert result.id.database is not None, f"database must be specified for
            # {self.id} or default_database set in config"
            filters = {}
            search_in_dbs = [self.id.database] if self.id.database else bw2data.databases
            for db in search_in_dbs:
                if self.id.location:
                    filters["location"] = self.id.location
                    search_results = bw2data.Database(db).search(self.id.name, filter=filters)
                else:
                    search_results = bw2data.Database(db).search(self.id.name)
                # print(len(search_results))
                # print(search_results)
                if self.id.unit:
                    search_results = list(filter(lambda a: a["unit"] == self.id.unit, search_results))
                #     if len(search_results) == 0:
                #         raise ValueError(f"No activity found with the specified unit {self.id}")
                # assert len(search_results) == 1, f"results : {len(search_results)}"
                if len(search_results) == 1:
                    bw_activity = search_results[0]
                    break
        if not bw_activity:
            raise ValueError(f"No activity found for {self.id}")
        self.id.fill_empty_fields(["name", "code", "location", "unit", ("alias", "name")],
                                  **bw_activity.as_dict())

        if not self.output:
            self.output = ActivityOutput(unit=bw_activity["unit"], magnitude=1.0)
        else:
            pass
        return ExtendedExperimentActivityData(**asdict(self),
                                              orig_id=orig_id,
                                              bw_activity=bw_activity)


@pydantic_dataclass(config=OperationConfig)
class ExtendedExperimentActivityData:
    id: ExperimentActivityId
    orig_id: ExperimentActivityId
    output: "ActivityOutput"
    bw_activity: Activity
    default_output_value: Optional[float] = 1.0  # this is the value converted to the default (bw) unit

    def __hash__(self):
        return self.bw_activity["code"]

    @property
    def alias(self):
        return self.id.alias


# TODO are we using this?
@pydantic_dataclass(frozen=True)
class BWMethod:
    description: str
    filename: str
    unit: str
    abbreviation: str
    num_cfs: int
    geocollections: list[str]


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentMethodData:
    id: tuple[str, ...]
    alias: Optional[str] = None

    @property
    def alias_(self) -> str:
        return str(self.alias)


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentMethodPrepData:
    id: tuple[str, ...]
    alias: str
    bw_method: BWMethod


@pydantic_dataclass(config=StrictInputConfig, repr=False)
class ExperimentScenarioData:
    # map from activity id to output. id is either as original (tuple) or alias-dict
    activities: Optional[Union[
        list[
            tuple[Union[str, ExperimentActivityId], ExperimentActivityOutput]],  # alias or id to output
        dict[str, ExperimentActivityOutput]]] = None  # alias to output, null means default-output (check exists)

    # either the alias, or the id of any method. not method means running them all
    methods: Optional[Union[list[Union[ExperimentMethodData, str]], dict[str, tuple[str, ...]]]] = None
    alias: Optional[str] = None

    @staticmethod
    def alias_factory(index: int):
        return f"Scenario {index}"


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentScenarioPrepData:
    activities: dict[SimpleScenarioActivityId, ExperimentActivityOutput] = Field(default_factory=dict)
    methods: list[ExperimentMethodData] = Field(default_factory=list)


@pydantic_dataclass(config=StrictInputConfig)
class ScenarioConfig:
    warn_default_demand: bool = True  # todo: bring this back
    include_bw_activity_in_nodes: bool = True
    store_raw_results: bool = False  # store numpy arrays of lca results
    # only used by ExperimentDataIO
    base_directory: Optional[str] = None
    # those are only used for testing
    debug_test_is_valid: bool = True
    debug_test_replace_bw_config: Union[bool, list[str]] = True  # standard replacement, or bw-project, bw-database
    debug_test_expected_error_code: Optional[int] = None
    debug_test_run: Optional[bool] = False
    note: Optional[str] = None


ActivitiesDataRows = list[ExperimentActivityData]
ActivitiesDataTypes = Union[ActivitiesDataRows, dict[str, ExperimentActivityData]]
MethodsDataTypes = Union[list[ExperimentMethodData], dict[str, tuple[str, ...]]]
HierarchyDataTypes = Union[list, dict]
ScenariosDataTypes = Union[list[ExperimentScenarioData], dict[str, ExperimentScenarioData]]


@pydantic_dataclass(config=StrictInputConfig)
class ExperimentData:
    """
    This class is used to store the data of an experiment.
    """
    bw_project: Union[str, EcoInventSimpleIndex] = Field(..., description="The brightway project name")
    activities: ActivitiesDataTypes = Field(..., description="The activities to be used in the experiment")
    methods: MethodsDataTypes = Field(..., description="The impact methods to be used in the experiment")
    bw_default_database: Optional[str] = Field(None,
                                               description="The default database of activities to be used "
                                                           "in the experiment")
    hierarchy: Optional[Union[list, dict]] = Field(None,
                                                   description="The activity hierarchy to be used in the experiment")
    scenarios: Optional[ScenariosDataTypes] = Field(None, description="The scenarios for this experiment")
    config: ScenarioConfig = Field(default_factory=ScenarioConfig,
                                   description="The configuration of this experiment")


@pydantic_dataclass
class ExperimentDataIO:
    bw_project: Union[str, EcoInventSimpleIndex]
    bw_default_database: Optional[str] = None
    activities: Optional[Union[ActivitiesDataTypes, str]] = None
    methods: Optional[Union[MethodsDataTypes, str]] = None
    hierarchy: Optional[Union[HierarchyDataTypes, str]] = None
    scenarios: Optional[Union[ScenariosDataTypes, str]] = None
    config: Optional[ScenarioConfig] = Field(default_factory=ScenarioConfig)


@dataclass
class BWCalculationSetup:
    name: str
    inv: list[dict[Activity, float]]
    ia: list[tuple[str, ...]]

    def register(self):
        bw2data.calculation_setups[self.name] = {
            "inv": self.inv,
            "ia": self.ia
        }


@dataclass
class ScenarioResultNodeData:
    output: tuple[Optional[str], Optional[float]] = (None, None)
    results: dict[str, float] = field(default_factory=dict)
    bw_activity: Optional[Activity] = None


Activity_Outputs = dict[SimpleScenarioActivityId, float]
